@@ -1,6 +1,6 @@
 
 /*
- * $Id: errorpage.c,v 1.186 2006/07/04 21:45:24 hno Exp $
+ * $Id: errorpage.c,v 1.193 2007/11/21 15:06:13 hno Exp $
  *
  * DEBUG: section 4     Error Generation
  * AUTHOR: Duane Wessels
@@ -261,13 +261,17 @@ errorPageName(int pageId)
  * Abstract:  This function creates a ErrorState object.
  */
 ErrorState *
-errorCon(err_type type, http_status status)
+errorCon(err_type type, http_status status, request_t * request)
 {
     ErrorState *err;
     err = cbdataAlloc(ErrorState);
     err->page_id = type;	/* has to be reset manually if needed */
     err->type = type;
     err->http_status = status;
+    if (request != NULL) {
+	err->request = requestLink(request);
+	err->src_addr = request->client_addr;
+    }
     return err;
 }
 
@@ -320,10 +324,10 @@ errorAppendEntry(StoreEntry * entry, ErrorState * err)
     authenticateFixHeader(rep, err->auth_user_request, err->request, 0, 1);
     httpReplySwapOut(rep, entry);
     EBIT_CLR(entry->flags, ENTRY_FWD_HDR_WAIT);
+    storeExpireNow(entry);
+    storeReleaseRequest(entry);
     storeBufferFlush(entry);
     storeComplete(entry);
-    storeNegativeCache(entry);
-    storeReleaseRequest(entry);
     storeUnlockObject(entry);
     errorStateFree(err);
 }
@@ -452,7 +456,7 @@ errorConvert(char token, ErrorState * err)
     memBufReset(&mb);
     switch (token) {
     case 'a':
-	if (r->auth_user_request)
+	if (r && r->auth_user_request)
 	    p = authenticateUserRequestUsername(r->auth_user_request);
 	if (!p)
 	    p = "-";
@@ -522,7 +526,7 @@ errorConvert(char token, ErrorState * err)
 	p = authenticateAuthUserRequestMessage(err->auth_user_request) ? authenticateAuthUserRequestMessage(err->auth_user_request) : "[not available]";
 	break;
     case 'M':
-	p = r ? RequestMethodStr[r->method] : "[unkown method]";
+	p = r ? RequestMethods[r->method].str : "[unknown method]";
 	break;
     case 'o':
 	p = external_acl_message;
@@ -543,7 +547,7 @@ errorConvert(char token, ErrorState * err)
 	if (NULL != r) {
 	    Packer p;
 	    memBufPrintf(&mb, "%s %s HTTP/%d.%d\n",
-		RequestMethodStr[r->method],
+		RequestMethods[r->method].str,
 		strLen(r->urlpath) ? strBuf(r->urlpath) : "/",
 		r->http_ver.major, r->http_ver.minor);
 	    packerToMemInit(&p, &mb);
@@ -621,20 +625,18 @@ errorBuildReply(ErrorState * err)
 {
     HttpReply *rep = httpReplyCreate();
     const char *name = errorPageName(err->page_id);
-    http_version_t version;
     /* no LMT for error pages; error pages expire immediately */
-    httpBuildVersion(&version, 1, 0);
     if (strchr(name, ':')) {
 	/* Redirection */
-	httpReplySetHeaders(rep, version, HTTP_MOVED_TEMPORARILY, NULL, "text/html", 0, 0, squid_curtime);
+	httpReplySetHeaders(rep, HTTP_MOVED_TEMPORARILY, NULL, "text/html", 0, -1, squid_curtime);
 	if (err->request) {
 	    char *quoted_url = rfc1738_escape_part(urlCanonical(err->request));
 	    httpHeaderPutStrf(&rep->header, HDR_LOCATION, name, quoted_url);
 	}
-	httpHeaderPutStrf(&rep->header, HDR_X_SQUID_ERROR, "%d %s\n", err->http_status, "Access Denied");
+	httpHeaderPutStrf(&rep->header, HDR_X_SQUID_ERROR, "%d %s", err->http_status, "Access Denied");
     } else {
 	MemBuf content = errorBuildContent(err);
-	httpReplySetHeaders(rep, version, err->http_status, NULL, "text/html", content.size, 0, squid_curtime);
+	httpReplySetHeaders(rep, err->http_status, NULL, "text/html", content.size, -1, squid_curtime);
 	/*
 	 * include some information for downstream caches. Implicit
 	 * replaceable content. This isn't quite sufficient. xerrno is not
