@@ -1,6 +1,6 @@
 
 /*
- * $Id: neighbors.c,v 1.299.2.2 2002/08/28 04:41:16 wessels Exp $
+ * $Id: neighbors.c,v 1.319.2.3 2008/05/04 23:23:13 hno Exp $
  *
  * DEBUG: section 15    Neighbor Routines
  * AUTHOR: Harvest Derived
@@ -38,7 +38,6 @@
 /* count mcast group peers every 15 minutes */
 #define MCAST_COUNT_RATE 900
 
-static int peerAllowedToUse(const peer *, request_t *);
 static int peerWouldBePinged(const peer *, request_t *);
 static void neighborRemove(peer *);
 static void neighborAlive(peer *, const MemObject *, const icp_common_t *);
@@ -48,8 +47,7 @@ static void neighborAliveHtcp(peer *, const MemObject *, const htcpReplyData *);
 static void neighborCountIgnored(peer *);
 static void peerRefreshDNS(void *);
 static IPH peerDNSConfigure;
-static void peerProbeConnect(peer *);
-static IPH peerProbeConnect2;
+static int peerProbeConnect(peer *);
 static CNCB peerProbeConnectDone;
 static void peerCountMcastPeersDone(void *data);
 static void peerCountMcastPeersStart(void *data);
@@ -106,6 +104,11 @@ neighborType(const peer * p, const request_t * request)
 	    if (d->type != PEER_NONE)
 		return d->type;
     }
+#if PEER_MULTICAST_SIBLINGS
+    if (p->type == PEER_MULTICAST)
+	if (p->options.mcast_siblings)
+	    return PEER_SIBLING;
+#endif
     return p->type;
 }
 
@@ -115,14 +118,19 @@ neighborType(const peer * p, const request_t * request)
  * this function figures out if it is appropriate to fetch REQUEST
  * from PEER.
  */
-static int
+int
 peerAllowedToUse(const peer * p, request_t * request)
 {
     const struct _domain_ping *d = NULL;
     int do_ping = 1;
-    aclCheck_t checklist;
     assert(request != NULL);
     if (neighborType(p, request) == PEER_SIBLING) {
+#if PEER_MULTICAST_SIBLINGS
+	if (p->type == PEER_MULTICAST && p->options.mcast_siblings &&
+	    (request->flags.nocache || request->flags.refresh || request->flags.loopdetect || request->flags.need_validation))
+	    debug(15, 2) ("peerAllowedToUse(%s, %s) : multicast-siblings optimization match\n",
+		p->name, request->host);
+#endif
 	if (request->flags.nocache)
 	    return 0;
 	if (request->flags.refresh)
@@ -146,21 +154,7 @@ peerAllowedToUse(const peer * p, request_t * request)
 	return do_ping;
     if (p->access == NULL)
 	return do_ping;
-    memset(&checklist, '\0', sizeof(checklist));
-    checklist.src_addr = request->client_addr;
-    checklist.my_addr = request->my_addr;
-    checklist.my_port = request->my_port;
-    checklist.request = request;
-#if 0 && USE_IDENT
-    /*
-     * this is currently broken because 'request->user_ident' has been
-     * moved to conn->rfc931 and we don't have access to the parent
-     * ConnStateData here.
-     */
-    if (request->user_ident[0])
-	xstrncpy(checklist.rfc931, request->user_ident, USER_IDENT_SZ);
-#endif
-    return aclCheckFast(p->access, &checklist);
+    return aclCheckFastRequest(p->access, request);
 }
 
 /* Return TRUE if it is okay to send an ICP request to this peer.   */
@@ -267,8 +261,12 @@ getRoundRobinParent(request_t * request)
 	    continue;
 	if (!peerHTTPOkay(p, request))
 	    continue;
-	if (q && q->rr_count < p->rr_count)
+	if (p->weight == 1) {
+	    if (q && q->rr_count < p->rr_count)
+		continue;
+	} else if (p->weight == 0 || (q && q->rr_count < (p->rr_count / p->weight))) {
 	    continue;
+	}
 	q = p;
     }
     if (q)
@@ -363,7 +361,7 @@ neighborRemove(peer * target)
 }
 
 void
-neighbors_open(int fd)
+neighbors_init(void)
 {
     struct sockaddr_in name;
     socklen_t len = sizeof(struct sockaddr_in);
@@ -371,25 +369,27 @@ neighbors_open(int fd)
     const char *me = getMyHostname();
     peer *this;
     peer *next;
-    memset(&name, '\0', sizeof(struct sockaddr_in));
-    if (getsockname(fd, (struct sockaddr *) &name, &len) < 0)
-	debug(15, 1) ("getsockname(%d,%p,%p) failed.\n", fd, &name, &len);
-    for (this = Config.peers; this; this = next) {
-	sockaddr_in_list *s;
-	next = this->next;
-	if (0 != strcmp(this->host, me))
-	    continue;
-	for (s = Config.Sockaddr.http; s; s = s->next) {
-	    if (this->http_port != ntohs(s->s.sin_port))
+    int fd = theInIcpConnection;
+    if (fd >= 0) {
+	memset(&name, '\0', sizeof(struct sockaddr_in));
+	if (getsockname(fd, (struct sockaddr *) &name, &len) < 0)
+	    debug(15, 1) ("getsockname(%d,%p,%p) failed.\n", fd, &name, &len);
+	for (this = Config.peers; this; this = next) {
+	    http_port_list *s;
+	    next = this->next;
+	    if (0 != strcmp(this->host, me))
 		continue;
-	    debug(15, 1) ("WARNING: Peer looks like this host\n");
-	    debug(15, 1) ("         Ignoring %s %s/%d/%d\n",
-		neighborTypeStr(this), this->host, this->http_port,
-		this->icp.port);
-	    neighborRemove(this);
+	    for (s = Config.Sockaddr.http; s; s = s->next) {
+		if (this->http_port != ntohs(s->s.sin_port))
+		    continue;
+		debug(15, 1) ("WARNING: Peer looks like this host\n");
+		debug(15, 1) ("         Ignoring %s %s/%d/%d\n",
+		    neighborTypeStr(this), this->host, this->http_port,
+		    this->icp.port);
+		neighborRemove(this);
+	    }
 	}
     }
-
     peerRefreshDNS((void *) 1);
     if (0 == echo_hdr.opcode) {
 	echo_hdr.opcode = ICP_SECHO;
@@ -406,9 +406,11 @@ neighbors_open(int fd)
     cachemgrRegister("server_list",
 	"Peer Cache Statistics",
 	neighborDumpPeers, 0, 1);
-    cachemgrRegister("non_peers",
-	"List of Unknown sites sending ICP messages",
-	neighborDumpNonPeers, 0, 1);
+    if (theInIcpConnection >= 0) {
+	cachemgrRegister("non_peers",
+	    "List of Unknown sites sending ICP messages",
+	    neighborDumpNonPeers, 0, 1);
+    }
 }
 
 int
@@ -430,6 +432,7 @@ neighborsUdpPing(request_t * request,
     int peers_pinged = 0;
     int parent_timeout = 0, parent_exprep = 0;
     int sibling_timeout = 0, sibling_exprep = 0;
+    int mcast_timeout = 0, mcast_exprep = 0;
 
     if (Config.peers == NULL)
 	return 0;
@@ -485,12 +488,8 @@ neighborsUdpPing(request_t * request,
 
 	p->stats.pings_sent++;
 	if (p->type == PEER_MULTICAST) {
-	    /*
-	     * set a bogus last_reply time so neighborUp() never
-	     * says a multicast peer is dead.
-	     */
-	    p->stats.last_reply = squid_curtime;
-	    (*exprep) += p->mcast.n_replies_expected;
+	    mcast_exprep += p->mcast.n_replies_expected;
+	    mcast_timeout += (p->stats.rtt * p->mcast.n_replies_expected);
 	} else if (neighborUp(p)) {
 	    /* its alive, expect a reply from it */
 	    if (neighborType(p, request) == PEER_PARENT) {
@@ -504,14 +503,17 @@ neighborsUdpPing(request_t * request,
 	    /* Neighbor is dead; ping it anyway, but don't expect a reply */
 	    /* log it once at the threshold */
 	    if (p->stats.logged_state == PEER_ALIVE) {
-		debug(15, 1) ("Detected DEAD %s: %s/%d/%d\n",
-		    neighborTypeStr(p),
-		    p->host, p->http_port, p->icp.port);
+		debug(15, 1) ("Detected DEAD %s: %s\n",
+		    neighborTypeStr(p), p->name);
 		p->stats.logged_state = PEER_DEAD;
 	    }
 	}
 	p->stats.last_query = squid_curtime;
-	if (p->stats.probe_start == 0)
+	/*
+	 * keep probe_start == 0 for a multicast peer,
+	 * so neighborUp() never says this peer is dead.
+	 */
+	if ((p->type != PEER_MULTICAST) && (p->stats.probe_start == 0))
 	    p->stats.probe_start = squid_curtime;
     }
     if ((first_ping = first_ping->next) == NULL)
@@ -551,7 +553,7 @@ neighborsUdpPing(request_t * request,
     /*
      * How many replies to expect?
      */
-    *exprep = parent_exprep + sibling_exprep;
+    *exprep = parent_exprep + sibling_exprep + mcast_exprep;
 
     /*
      * If there is a configured timeout, use it
@@ -562,6 +564,8 @@ neighborsUdpPing(request_t * request,
 	if (*exprep > 0) {
 	    if (parent_exprep)
 		*timeout = 2 * parent_timeout / parent_exprep;
+	    else if (mcast_exprep)
+		*timeout = 2 * mcast_timeout / mcast_exprep;
 	    else
 		*timeout = 2 * sibling_timeout / sibling_exprep;
 	} else
@@ -569,6 +573,9 @@ neighborsUdpPing(request_t * request,
 	if (Config.Timeout.icp_query_max)
 	    if (*timeout > Config.Timeout.icp_query_max)
 		*timeout = Config.Timeout.icp_query_max;
+
+	if (*timeout < Config.Timeout.icp_query_min)
+	    *timeout = Config.Timeout.icp_query_min;
     }
     return peers_pinged;
 }
@@ -589,14 +596,11 @@ peerDigestLookup(peer * p, request_t * request)
     } else if (!peerHTTPOkay(p, request)) {
 	debug(15, 5) ("peerDigestLookup: !peerHTTPOkay\n");
 	return LOOKUP_NONE;
-    } else if (p->digest->flags.usable) {
-	debug(15, 5) ("peerDigestLookup: usable\n");
-	/* fall through; put here to have common case on top */ ;
     } else if (!p->digest->flags.needed) {
 	debug(15, 5) ("peerDigestLookup: note need\n");
 	peerDigestNeeded(p->digest);
 	return LOOKUP_NONE;
-    } else {
+    } else if (!p->digest->flags.usable) {
 	debug(15, 5) ("peerDigestLookup: !ready && %srequested\n",
 	    p->digest->flags.requested ? "" : "!");
 	return LOOKUP_NONE;
@@ -681,9 +685,8 @@ static void
 neighborAlive(peer * p, const MemObject * mem, const icp_common_t * header)
 {
     if (p->stats.logged_state == PEER_DEAD && p->tcp_up) {
-	debug(15, 1) ("Detected REVIVED %s: %s/%d/%d\n",
-	    neighborTypeStr(p),
-	    p->host, p->http_port, p->icp.port);
+	debug(15, 1) ("Detected REVIVED %s: %s\n",
+	    neighborTypeStr(p), p->name);
 	p->stats.logged_state = PEER_ALIVE;
     }
     p->stats.last_reply = squid_curtime;
@@ -714,9 +717,8 @@ static void
 neighborAliveHtcp(peer * p, const MemObject * mem, const htcpReplyData * htcp)
 {
     if (p->stats.logged_state == PEER_DEAD && p->tcp_up) {
-	debug(15, 1) ("Detected REVIVED %s: %s/%d/%d\n",
-	    neighborTypeStr(p),
-	    p->host, p->http_port, p->icp.port);
+	debug(15, 1) ("Detected REVIVED %s: %s\n",
+	    neighborTypeStr(p), p->name);
 	p->stats.logged_state = PEER_ALIVE;
     }
     p->stats.last_reply = squid_curtime;
@@ -796,7 +798,7 @@ neighborsUdpAck(const cache_key * key, icp_common_t * header, const struct socka
     StoreEntry *entry;
     MemObject *mem = NULL;
     peer_t ntype = PEER_NONE;
-    char *opcode_d;
+    const char *opcode_d;
     icp_opcode opcode = (icp_opcode) header->opcode;
 
     debug(15, 6) ("neighborsUdpAck: opcode %d '%s'\n",
@@ -907,7 +909,7 @@ peerFindByName(const char *name)
 {
     peer *p = NULL;
     for (p = Config.peers; p; p = p->next) {
-	if (!strcasecmp(name, p->host))
+	if (!strcasecmp(name, p->name))
 	    break;
     }
     return p;
@@ -918,7 +920,7 @@ peerFindByNameAndPort(const char *name, unsigned short port)
 {
     peer *p = NULL;
     for (p = Config.peers; p; p = p->next) {
-	if (strcasecmp(name, p->host))
+	if (strcasecmp(name, p->name))
 	    continue;
 	if (port != p->http_port)
 	    continue;
@@ -931,13 +933,12 @@ int
 neighborUp(const peer * p)
 {
     if (!p->tcp_up) {
-	peerProbeConnect((peer *) p);
-	return 0;
+	if (!peerProbeConnect((peer *) p))
+	    return 0;
     }
-    if (p->options.no_query)
-	return 1;
-    if (p->stats.probe_start != 0 &&
-	squid_curtime - p->stats.probe_start > Config.Timeout.deadPeer)
+    if (p->stats.logged_state != PEER_ALIVE)
+	return 0;
+    if (p->monitor.state != PEER_ALIVE)
 	return 0;
     /*
      * The peer can not be UP if we don't have any IP addresses
@@ -945,6 +946,11 @@ neighborUp(const peer * p)
      */
     if (0 == p->n_addresses)
 	return 0;
+    if (p->options.no_query)
+	return 1;
+    if (p->stats.probe_start != 0 &&
+	squid_curtime - p->stats.probe_start > Config.Timeout.deadPeer)
+	return 0;
     return 1;
 }
 
@@ -961,7 +967,10 @@ peerDestroy(void *data)
 	safe_free(l->domain);
 	safe_free(l);
     }
+    aclDestroyAccessList(&p->access);
     safe_free(p->host);
+    safe_free(p->name);
+    safe_free(p->domain);
 #if USE_CACHE_DIGESTS
     if (p->digest) {
 	PeerDigest *pd = p->digest;
@@ -1009,6 +1018,8 @@ peerDNSConfigure(const ipcache_addrs * ia, void *data)
 	debug(15, 2) ("--> IP address #%d: %s\n", j, inet_ntoa(p->addresses[j]));
 	p->n_addresses++;
     }
+    if (!p->tcp_up)
+	peerProbeConnect((peer *) p);
     ap = &p->in_addr;
     memset(ap, '\0', sizeof(struct sockaddr_in));
     ap->sin_family = AF_INET;
@@ -1032,72 +1043,87 @@ peerRefreshDNS(void *data)
 	eventAddIsh("peerRefreshDNS", peerRefreshDNS, NULL, 180.0, 1);
 	return;
     }
-    for (p = Config.peers; p; p = p->next)
+    for (p = Config.peers; p; p = p->next) {
 	ipcache_nbgethostbyname(p->host, peerDNSConfigure, p);
+    }
     /* Reconfigure the peers every hour */
     eventAddIsh("peerRefreshDNS", peerRefreshDNS, NULL, 3600.0, 1);
 }
 
-void
-peerConnectFailed(peer * p)
+static void
+peerConnectFailedSilent(peer * p)
 {
     p->stats.last_connect_failure = squid_curtime;
     if (!p->tcp_up) {
 	debug(15, 2) ("TCP connection to %s/%d dead\n", p->host, p->http_port);
 	return;
     }
-    debug(15, 1) ("TCP connection to %s/%d failed\n", p->host, p->http_port);
     p->tcp_up--;
     if (!p->tcp_up) {
-	debug(15, 1) ("Detected DEAD %s: %s/%d/%d\n",
-	    neighborTypeStr(p),
-	    p->host, p->http_port, p->icp.port);
+	debug(15, 1) ("Detected DEAD %s: %s\n",
+	    neighborTypeStr(p), p->name);
 	p->stats.logged_state = PEER_DEAD;
     }
 }
 
+void
+peerConnectFailed(peer * p)
+{
+    debug(15, 1) ("TCP connection to %s/%d failed\n", p->host, p->http_port);
+    peerConnectFailedSilent(p);
+}
+
 void
 peerConnectSucceded(peer * p)
 {
     if (!p->tcp_up) {
 	debug(15, 2) ("TCP connection to %s/%d succeded\n", p->host, p->http_port);
-	debug(15, 1) ("Detected REVIVED %s: %s/%d/%d\n",
-	    neighborTypeStr(p),
-	    p->host, p->http_port, p->icp.port);
+	debug(15, 1) ("Detected REVIVED %s: %s\n",
+	    neighborTypeStr(p), p->name);
+	peerMonitorNow(p);
 	p->stats.logged_state = PEER_ALIVE;
+	if (!p->n_addresses)
+	    ipcache_nbgethostbyname(p->host, peerDNSConfigure, p);
     }
     p->tcp_up = PEER_TCP_MAGIC_COUNT;
 }
 
+static void
+peerProbeConnectTimeout(int fd, void *data)
+{
+    peer *p = data;
+    comm_close(fd);
+    p->test_fd = -1;
+    peerConnectFailedSilent(p);
+}
+
 /*
  * peerProbeConnect will be called on dead peers by neighborUp 
  */
-static void
+static int
 peerProbeConnect(peer * p)
 {
     int fd;
+    time_t ctimeout = p->connect_timeout > 0 ? p->connect_timeout
+    : Config.Timeout.peer_connect;
+    int ret = squid_curtime - p->stats.last_connect_failure > ctimeout * 10;
     if (p->test_fd != -1)
-	return;			/* probe already running */
-    if (squid_curtime - p->stats.last_connect_probe < Config.Timeout.connect)
-	return;			/* don't probe to often */
-    fd = comm_open(SOCK_STREAM, 0, getOutgoingAddr(NULL),
+	return ret;		/* probe already running */
+    if (squid_curtime - p->stats.last_connect_probe == 0)
+	return ret;		/* don't probe to often */
+    fd = comm_open(SOCK_STREAM, IPPROTO_TCP, getOutgoingAddr(NULL),
 	0, COMM_NONBLOCKING, p->host);
     if (fd < 0)
-	return;
+	return ret;
+    commSetTimeout(fd, ctimeout, peerProbeConnectTimeout, p);
     p->test_fd = fd;
     p->stats.last_connect_probe = squid_curtime;
-    ipcache_nbgethostbyname(p->host, peerProbeConnect2, p);
-}
-
-static void
-peerProbeConnect2(const ipcache_addrs * ianotused, void *data)
-{
-    peer *p = data;
     commConnectStart(p->test_fd,
 	p->host,
 	p->http_port,
 	peerProbeConnectDone,
 	p);
+    return ret;
 }
 
 static void
@@ -1107,7 +1133,7 @@ peerProbeConnectDone(int fd, int status, void *data)
     if (status == COMM_OK) {
 	peerConnectSucceded(p);
     } else {
-	peerConnectFailed(p);
+	peerConnectFailedSilent(p);
     }
     comm_close(fd);
     p->test_fd = -1;
@@ -1139,12 +1165,13 @@ peerCountMcastPeersStart(void *data)
     assert(p->type == PEER_MULTICAST);
     p->mcast.flags.count_event_pending = 0;
     snprintf(url, MAX_URL, "http://%s/", inet_ntoa(p->in_addr.sin_addr));
-    fake = storeCreateEntry(url, url, null_request_flags, METHOD_GET);
+    fake = storeCreateEntry(url, null_request_flags, METHOD_GET);
     psstate = cbdataAlloc(ps_state);
     psstate->request = requestLink(urlParse(METHOD_GET, url));
     psstate->entry = fake;
     psstate->callback = NULL;
     psstate->callback_data = p;
+    cbdataLock(psstate->callback_data);
     psstate->ping.start = current_time;
     mem = fake->mem_obj;
     mem->request = requestLink(psstate->request);
@@ -1164,7 +1191,7 @@ peerCountMcastPeersStart(void *data)
     eventAdd("peerCountMcastPeersDone",
 	peerCountMcastPeersDone,
 	psstate,
-	(double) Config.Timeout.mcast_icp_query, 1);
+	Config.Timeout.mcast_icp_query / 1000.0, 1);
     p->mcast.flags.counting = 1;
     peerCountMcastPeersSchedule(p, MCAST_COUNT_RATE);
 }
@@ -1175,17 +1202,20 @@ peerCountMcastPeersDone(void *data)
     ps_state *psstate = data;
     peer *p = psstate->callback_data;
     StoreEntry *fake = psstate->entry;
-    p->mcast.flags.counting = 0;
-    p->mcast.avg_n_members = doubleAverage(p->mcast.avg_n_members,
-	(double) psstate->ping.n_recv,
-	++p->mcast.n_times_counted,
-	10);
-    debug(15, 1) ("Group %s: %d replies, %4.1f average, RTT %d\n",
-	p->host,
-	psstate->ping.n_recv,
-	p->mcast.avg_n_members,
-	p->stats.rtt);
-    p->mcast.n_replies_expected = (int) p->mcast.avg_n_members;
+    if (cbdataValid(p)) {
+	p->mcast.flags.counting = 0;
+	p->mcast.avg_n_members = doubleAverage(p->mcast.avg_n_members,
+	    (double) psstate->ping.n_recv,
+	    ++p->mcast.n_times_counted,
+	    10);
+	debug(15, 1) ("Group %s: %d replies, %4.1f average, RTT %d\n",
+	    p->host,
+	    psstate->ping.n_recv,
+	    p->mcast.avg_n_members,
+	    p->stats.rtt);
+	p->mcast.n_replies_expected = (int) p->mcast.avg_n_members;
+    }
+    cbdataUnlock(p);
     EBIT_SET(fake->flags, ENTRY_ABORTED);
     requestUnlink(fake->mem_obj->request);
     fake->mem_obj->request = NULL;
@@ -1236,8 +1266,18 @@ dump_peer_options(StoreEntry * sentry, peer * p)
 	storeAppendPrintf(sentry, " round-robin");
     if (p->options.mcast_responder)
 	storeAppendPrintf(sentry, " multicast-responder");
+#if PEER_MULTICAST_SIBLINGS
+    if (p->options.mcast_siblings)
+	storeAppendPrintf(sentry, " multicast-siblings");
+#endif
+    if (p->weight != 1)
+	storeAppendPrintf(sentry, " weight=%d", p->weight);
     if (p->options.closest_only)
 	storeAppendPrintf(sentry, " closest-only");
+    if (p->options.userhash)
+	storeAppendPrintf(sentry, " userhash");
+    if (p->options.sourcehash)
+	storeAppendPrintf(sentry, " sourcehash");
 #if USE_HTCP
     if (p->options.htcp)
 	storeAppendPrintf(sentry, " htcp");
@@ -1252,6 +1292,29 @@ dump_peer_options(StoreEntry * sentry, peer * p)
 	storeAppendPrintf(sentry, " login=%s", p->login);
     if (p->mcast.ttl > 0)
 	storeAppendPrintf(sentry, " ttl=%d", p->mcast.ttl);
+    if (p->connect_timeout > 0)
+	storeAppendPrintf(sentry, " connect-timeout=%d", (int) p->connect_timeout);
+#if USE_CACHE_DIGESTS
+    if (p->digest_url)
+	storeAppendPrintf(sentry, " digest-url=%s", p->digest_url);
+#endif
+    if (p->options.allow_miss)
+	storeAppendPrintf(sentry, " allow-miss");
+    if (p->max_conn > 0)
+	storeAppendPrintf(sentry, " max-conn=%d", p->max_conn);
+    if (p->options.originserver)
+	storeAppendPrintf(sentry, " originserver");
+    /* name is used in the heading */
+    if (p->monitor.url)
+	storeAppendPrintf(sentry, " monitorurl=%s", p->monitor.url);
+    if (p->monitor.min > 0 || p->monitor.max > 0) {
+	if (p->monitor.max > 0)
+	    storeAppendPrintf(sentry, " monitorsize=%d:%d", p->monitor.min, p->monitor.max);
+	else
+	    storeAppendPrintf(sentry, " monitorsize=%d", p->monitor.min);
+    }
+    if (p->domain)
+	storeAppendPrintf(sentry, " forceddomain=%s", p->domain);
     storeAppendPrintf(sentry, "\n");
 }
 
@@ -1266,8 +1329,10 @@ dump_peers(StoreEntry * sentry, peer * peers)
 	storeAppendPrintf(sentry, "There are no neighbors installed.\n");
     for (e = peers; e; e = e->next) {
 	assert(e->host != NULL);
-	storeAppendPrintf(sentry, "\n%-11.11s: %s/%d/%d\n",
+	storeAppendPrintf(sentry, "\n%-11.11s: %s\n",
 	    neighborTypeStr(e),
+	    e->name);
+	storeAppendPrintf(sentry, "Host       : %s/%d/%d\n",
 	    e->host,
 	    e->http_port,
 	    e->icp.port);
@@ -1281,42 +1346,49 @@ dump_peers(StoreEntry * sentry, peer * peers)
 	    neighborUp(e) ? "Up" : "Down");
 	storeAppendPrintf(sentry, "AVG RTT    : %d msec\n", e->stats.rtt);
 	storeAppendPrintf(sentry, "OPEN CONNS : %d\n", e->stats.conn_open);
-	storeAppendPrintf(sentry, "LAST QUERY : %8d seconds ago\n",
-	    (int) (squid_curtime - e->stats.last_query));
-	storeAppendPrintf(sentry, "LAST REPLY : %8d seconds ago\n",
-	    (int) (squid_curtime - e->stats.last_reply));
-	storeAppendPrintf(sentry, "PINGS SENT : %8d\n", e->stats.pings_sent);
-	storeAppendPrintf(sentry, "PINGS ACKED: %8d %3d%%\n",
-	    e->stats.pings_acked,
-	    percent(e->stats.pings_acked, e->stats.pings_sent));
+	if (!e->options.no_query) {
+	    storeAppendPrintf(sentry, "LAST QUERY : %8d seconds ago\n",
+		(int) (squid_curtime - e->stats.last_query));
+	    if (e->stats.last_reply > 0)
+		storeAppendPrintf(sentry, "LAST REPLY : %8d seconds ago\n",
+		    (int) (squid_curtime - e->stats.last_reply));
+	    else
+		storeAppendPrintf(sentry, "LAST REPLY : none received\n");
+	    storeAppendPrintf(sentry, "PINGS SENT : %8d\n", e->stats.pings_sent);
+	    storeAppendPrintf(sentry, "PINGS ACKED: %8d %3d%%\n",
+		e->stats.pings_acked,
+		percent(e->stats.pings_acked, e->stats.pings_sent));
+	}
 	storeAppendPrintf(sentry, "FETCHES    : %8d %3d%%\n",
 	    e->stats.fetches,
 	    percent(e->stats.fetches, e->stats.pings_acked));
 	storeAppendPrintf(sentry, "IGNORED    : %8d %3d%%\n",
 	    e->stats.ignored_replies,
 	    percent(e->stats.ignored_replies, e->stats.pings_acked));
-	storeAppendPrintf(sentry, "Histogram of PINGS ACKED:\n");
+	if (!e->options.no_query) {
+	    storeAppendPrintf(sentry, "Histogram of PINGS ACKED:\n");
 #if USE_HTCP
-	if (e->options.htcp) {
-	    storeAppendPrintf(sentry, "\tMisses\t%8d %3d%%\n",
-		e->htcp.counts[0],
-		percent(e->htcp.counts[0], e->stats.pings_acked));
-	    storeAppendPrintf(sentry, "\tHits\t%8d %3d%%\n",
-		e->htcp.counts[1],
-		percent(e->htcp.counts[1], e->stats.pings_acked));
-	} else {
+	    if (e->options.htcp) {
+		storeAppendPrintf(sentry, "\tMisses\t%8d %3d%%\n",
+		    e->htcp.counts[0],
+		    percent(e->htcp.counts[0], e->stats.pings_acked));
+		storeAppendPrintf(sentry, "\tHits\t%8d %3d%%\n",
+		    e->htcp.counts[1],
+		    percent(e->htcp.counts[1], e->stats.pings_acked));
+	    } else {
 #endif
-	    for (op = ICP_INVALID; op < ICP_END; op++) {
-		if (e->icp.counts[op] == 0)
-		    continue;
-		storeAppendPrintf(sentry, "    %12.12s : %8d %3d%%\n",
-		    icp_opcode_str[op],
-		    e->icp.counts[op],
-		    percent(e->icp.counts[op], e->stats.pings_acked));
+		for (op = ICP_INVALID; op < ICP_END; op++) {
+		    if (e->icp.counts[op] == 0)
+			continue;
+		    storeAppendPrintf(sentry, "    %12.12s : %8d %3d%%\n",
+			icp_opcode_str[op],
+			e->icp.counts[op],
+			percent(e->icp.counts[op], e->stats.pings_acked));
+		}
+#if USE_HTCP
 	    }
-#if USE_HTCP
-	}
 #endif
+	}
 	if (e->stats.last_connect_failure) {
 	    storeAppendPrintf(sentry, "Last failed connect() at: %s\n",
 		mkhttpdlogtime(&(e->stats.last_connect_failure)));
@@ -1350,14 +1422,14 @@ neighborsHtcpReply(const cache_key * key, htcpReplyData * htcp, const struct soc
 	neighborAliveHtcp(p, mem, htcp);
     /* Does the entry exist? */
     if (NULL == e) {
-	debug(12, 3) ("neighyborsHtcpReply: Cache key '%s' not found\n",
+	debug(12, 3) ("neighborsHtcpReply: Cache key '%s' not found\n",
 	    storeKeyText(key));
 	neighborCountIgnored(p);
 	return;
     }
     /* check if someone is already fetching it */
     if (EBIT_TEST(e->flags, ENTRY_DISPATCHED)) {
-	debug(15, 3) ("neighborsUdpAck: '%s' already being fetched.\n",
+	debug(15, 3) ("neighborsHtcpReply: '%s' already being fetched.\n",
 	    storeKeyText(key));
 	neighborCountIgnored(p);
 	return;
@@ -1369,13 +1441,13 @@ neighborsHtcpReply(const cache_key * key, htcpReplyData * htcp, const struct soc
 	return;
     }
     if (e->ping_status != PING_WAITING) {
-	debug(15, 2) ("neighborsUdpAck: Entry %s is not PING_WAITING\n",
+	debug(15, 2) ("neighborsHtcpReply: Entry %s is not PING_WAITING\n",
 	    storeKeyText(key));
 	neighborCountIgnored(p);
 	return;
     }
     if (e->lock_count == 0) {
-	debug(12, 1) ("neighborsUdpAck: '%s' has no locks\n",
+	debug(12, 1) ("neighborsHtcpReply: '%s' has no locks\n",
 	    storeKeyText(key));
 	neighborCountIgnored(p);
 	return;
