@@ -1,6 +1,6 @@
 
 /*
- * $Id: protos.h,v 1.507 2006/07/30 23:27:03 hno Exp $
+ * $Id: protos.h,v 1.547.2.7 2008/05/27 12:49:39 hno Exp $
  *
  *
  * SQUID Web Proxy Cache          http://www.squid-cache.org/
@@ -56,8 +56,10 @@ char *log_quote(const char *header);
 extern aclCheck_t *aclChecklistCreate(const struct _acl_access *,
     request_t *,
     const char *ident);
+void aclChecklistCacheInit(aclCheck_t * checklist);
 extern void aclNBCheck(aclCheck_t *, PF *, void *);
 extern int aclCheckFast(const struct _acl_access *A, aclCheck_t *);
+int aclCheckFastRequest(const acl_access * A, request_t * request);
 extern void aclChecklistFree(aclCheck_t *);
 extern int aclMatchAclList(const acl_list * list, aclCheck_t * checklist);
 extern void aclDestroyAccessList(struct _acl_access **list);
@@ -67,7 +69,7 @@ extern void aclParseAccessLine(struct _acl_access **);
 extern void aclParseAclList(acl_list **);
 extern void aclParseAclLine(acl **);
 extern int aclIsProxyAuth(const char *name);
-extern err_type aclGetDenyInfoPage(acl_deny_info_list ** head, const char *name);
+extern err_type aclGetDenyInfoPage(acl_deny_info_list ** head, const char *name, int redirect_allowed);
 extern void aclParseDenyInfoLine(struct _acl_deny_info_list **);
 extern void aclDestroyDenyInfoList(struct _acl_deny_info_list **);
 extern void aclDestroyRegexList(struct _relist *data);
@@ -95,6 +97,8 @@ extern void wordlistCat(const wordlist *, MemBuf * mb);
 extern void allocate_new_swapdir(cacheSwap *);
 extern void self_destruct(void);
 extern int GetInteger(void);
+extern int xatoi(const char *);
+extern unsigned short xatos(const char *);
 
 /* extra functions from cache_cf.c useful for lib modules */
 extern void parse_int(int *var);
@@ -143,10 +147,15 @@ extern int isTcpHit(log_type);
 extern void clientPinConnection(ConnStateData * conn, int fd, const request_t * request, peer * peer, int auth);
 extern int clientGetPinnedInfo(const ConnStateData * conn, const request_t * request, peer ** peer);
 extern int clientGetPinnedConnection(ConnStateData * conn, const request_t * request, const peer * peer, int *auth);
+extern void clientReassignDelaypools(void);
 
 extern int commSetNonBlocking(int fd);
 extern int commUnsetNonBlocking(int fd);
 extern void commSetCloseOnExec(int fd);
+extern void commSetTcpKeepalive(int fd, int idle, int interval, int timeout);
+extern int commSetTos(int fd, int tos);
+extern int commSetSocketPriority(int fd, int prio);
+extern int commSetIPOption(int fd, uint8_t option, void *value, size_t size);
 extern int comm_accept(int fd, struct sockaddr_in *, struct sockaddr_in *);
 extern void comm_close(int fd);
 extern void comm_reset_close(int fd);
@@ -159,11 +168,14 @@ extern void comm_init(void);
 extern int comm_listen(int sock);
 extern int comm_open(int, int, struct in_addr, u_short port, int, const char *note);
 extern int comm_openex(int, int, struct in_addr, u_short, int, unsigned char TOS, const char *);
+extern int comm_fdopen(int, int, struct in_addr, u_short, int, const char *);
+extern int comm_fdopenex(int, int, struct in_addr, u_short, int, unsigned char, const char *);
 extern u_short comm_local_port(int fd);
 
 extern void commDeferFD(int fd);
 extern void commResumeFD(int fd);
 extern void commSetSelect(int, unsigned int, PF *, void *, time_t);
+extern void commRemoveSlow(int fd);
 extern void comm_add_close_handler(int fd, PF *, void *);
 extern void comm_remove_close_handler(int fd, PF *, void *);
 extern int comm_udp_sendto(int, const struct sockaddr_in *, int, const void *, int);
@@ -174,6 +186,15 @@ extern void comm_write(int fd,
     void *handler_data,
     FREE *);
 extern void comm_write_mbuf(int fd, MemBuf mb, CWCB * handler, void *handler_data);
+extern void comm_write_header(int fd,
+    const char *buf,
+    int size,
+    const char *header,
+    size_t header_size,
+    CWCB * handler,
+    void *handler_data,
+    FREE *);
+extern void comm_write_mbuf_header(int fd, MemBuf mb, const char *header, size_t header_size, CWCB * handler, void *handler_data);
 extern void commCallCloseHandlers(int fd);
 extern int commSetTimeout(int fd, int, PF *, void *);
 extern void commSetDefer(int fd, DEFER * func, void *);
@@ -185,13 +206,17 @@ extern void commCloseAllSockets(void);
  * comm_select.c
  */
 extern void comm_select_init(void);
+extern void comm_select_postinit(void);
 extern void comm_select_shutdown(void);
 extern int comm_select(int);
 extern void commUpdateEvents(int fd);
 extern void commSetEvents(int fd, int need_read, int need_write);
+extern void commClose(int fd);
+extern void commOpen(int fd);
 extern void commUpdateReadHandler(int, PF *, void *);
 extern void commUpdateWriteHandler(int, PF *, void *);
 extern void comm_quick_poll_required(void);
+extern void comm_select_status(StoreEntry *);
 
 extern void packerToStoreInit(Packer * p, StoreEntry * e);
 extern void packerToMemInit(Packer * p, MemBuf * mb);
@@ -219,6 +244,7 @@ _db_print(const char *,...) PRINTF_FORMAT_ARG1;
 #else
 extern void _db_print();
 #endif
+extern int debug_log_flush(void);
 extern void xassert(const char *, const char *, int);
 
 /* packs, then prints an object using debug() */
@@ -253,9 +279,11 @@ extern void eventCleanup(void);
 extern void eventFreeMemory(void);
 extern int eventFind(EVH *, void *);
 
+extern void fd_init(void);
 extern void fd_close(int fd);
 extern void fd_open(int fd, unsigned int type, const char *);
 extern void fd_note(int fd, const char *);
+extern void fd_note_static(int fd, const char *);
 extern void fd_bytes(int fd, int len, unsigned int type);
 extern void fdFreeMemory(void);
 extern void fdDumpOpen(void);
@@ -393,12 +421,13 @@ extern void httpHeaderAddContRange(HttpHeader *, HttpHdrRangeSpec, squid_off_t);
 extern void strListAdd(String * str, const char *item, char del);
 extern void strListAddUnique(String * str, const char *item, char del);
 extern int strListIsMember(const String * str, const char *item, char del);
-extern int strListIsSubstr(const String * list, const char *s, char del);
+extern int strIsSubstr(const String * list, const char *s);
 extern int strListGetItem(const String * str, char del, const char **item, int *ilen, const char **pos);
 extern const char *getStringPrefix(const char *str, const char *end);
 extern int httpHeaderParseInt(const char *start, int *val);
 extern int httpHeaderParseSize(const char *start, squid_off_t * sz);
 extern int httpHeaderReset(HttpHeader * hdr);
+extern void httpHeaderAddClone(HttpHeader * hdr, const HttpHeaderEntry * e);
 #if STDC_HEADERS
 extern void
 httpHeaderPutStrf(HttpHeader * hdr, http_hdr_type id, const char *fmt,...) PRINTF_FORMAT_ARG3;
@@ -463,7 +492,7 @@ extern void httpHdrMangleList(HttpHeader *, request_t *);
 
 /* Http Msg (currently in HttpReply.c @?@ ) */
 extern int httpMsgIsPersistent(http_version_t http_ver, const HttpHeader * hdr);
-extern int httpMsgIsolateHeaders(const char **parse_start, const char **blk_start, const char **blk_end);
+extern int httpMsgIsolateHeaders(const char **parse_start, int l, const char **blk_start, const char **blk_end);
 
 /* Http Reply */
 extern void httpReplyInitModule(void);
@@ -481,13 +510,13 @@ extern MemBuf httpReplyPack(const HttpReply * rep);
 /* swap: create swap-based packer, pack, destroy packer and absorbs the reply if not the same as the object reply */
 extern void httpReplySwapOut(HttpReply * rep, StoreEntry * e);
 /* set commonly used info with one call */
-extern void httpReplySetHeaders(HttpReply * rep, http_version_t ver, http_status status,
+extern void httpReplySetHeaders(HttpReply * rep, http_status status,
     const char *reason, const char *ctype, squid_off_t clen, time_t lmt, time_t expires);
 /* do everything in one call: init, set, pack, clean, return MemBuf */
 extern MemBuf httpPackedReply(http_version_t ver, http_status status, const char *ctype,
     squid_off_t clen, time_t lmt, time_t expires);
 /* construct 304 reply and pack it into MemBuf, return MemBuf */
-extern MemBuf httpPacked304Reply(const HttpReply * rep);
+extern MemBuf httpPacked304Reply(const HttpReply * rep, int http11);
 /* update when 304 reply is received for a cached object */
 extern void httpReplyUpdateOnNotModified(HttpReply * rep, HttpReply * freshRep);
 /* header manipulation */
@@ -497,13 +526,13 @@ extern time_t httpReplyExpires(const HttpReply * rep);
 extern int httpReplyHasCc(const HttpReply * rep, http_hdr_cc_type type);
 extern void httpRedirectReply(HttpReply *, http_status, const char *);
 extern squid_off_t httpReplyBodySize(method_t, const HttpReply *);
+extern HttpReply *httpReplyClone(HttpReply * src);
 
 /* Http Request */
 extern request_t *requestCreate(method_t, protocol_t, const char *urlpath);
 extern void requestDestroy(request_t *);
 extern request_t *requestLink(request_t *);
 extern void requestUnlink(request_t *);
-extern int httpRequestParseHeader(request_t * req, const char *parse_start);
 extern void httpRequestSwapOut(const request_t * req, StoreEntry * e);
 extern void httpRequestPackDebug(request_t * req, Packer * p);
 extern int httpRequestPrefixLen(const request_t * req);
@@ -603,7 +632,7 @@ extern void memBufReset(MemBuf * mb);
 /* unfirtunate hack to test if the buffer has been Init()ialized */
 extern int memBufIsNull(MemBuf * mb);
 /* calls memcpy, appends exactly size bytes, extends buffer if needed */
-extern void memBufAppend(MemBuf * mb, const char *buf, int size);
+extern void memBufAppend(MemBuf * mb, const void *buf, int size);
 /* calls snprintf, extends buffer if needed */
 #if STDC_HEADERS
 extern void
@@ -683,7 +712,6 @@ extern int peerAllowedToUse(const peer *, request_t *);
 extern void netdbInit(void);
 extern void netdbHandlePingReply(const struct sockaddr_in *from, int hops, int rtt);
 extern void netdbPingSite(const char *hostname);
-extern void netdbDump(StoreEntry *);
 extern int netdbHops(struct in_addr);
 extern void netdbFreeMemory(void);
 extern int netdbHostHops(const char *host);
@@ -733,6 +761,10 @@ extern void redirectStart(clientHttpRequest *, RH *, void *);
 extern void redirectInit(void);
 extern void redirectShutdown(void);
 
+extern void storeurlStart(clientHttpRequest *, RH *, void *);
+extern void storeurlInit(void);
+extern void storeurlShutdown(void);
+
 extern void locationRewriteStart(HttpReply *, clientHttpRequest *, RH *, void *);
 extern void locationRewriteInit(void);
 extern void locationRewriteShutdown(void);
@@ -782,9 +814,11 @@ extern void refreshAddToList(const char *, int, time_t, int, time_t);
 extern int refreshIsCachable(const StoreEntry *);
 extern int refreshCheckHTTP(const StoreEntry *, request_t *);
 extern int refreshCheckHTTPStale(const StoreEntry *, request_t *);
+extern int refreshCheckStaleOK(const StoreEntry *, request_t *);
 extern int refreshCheckICP(const StoreEntry *, request_t *);
 extern int refreshCheckHTCP(const StoreEntry *, request_t *);
 extern int refreshCheckDigest(const StoreEntry *, time_t delta);
+extern refresh_cc refreshCC(const StoreEntry *, request_t *);
 extern time_t getMaxAge(const char *url);
 extern void refreshInit(void);
 extern const refresh_t *refreshLimits(const char *url);
@@ -797,7 +831,6 @@ extern void reconfigure(int);
 
 extern void start_announce(void *unused);
 extern void sslStart(clientHttpRequest *, squid_off_t *, int *);
-extern void waisStart(FwdState *);
 
 /* ident.c */
 #if USE_IDENT
@@ -859,13 +892,14 @@ extern void memFreeString(size_t size, void *);
 extern void memFreeBuf(size_t size, void *);
 extern FREE *memFreeBufFunc(size_t size);
 extern int memInUse(mem_type);
-extern int memTotalAllocated(void);
+extern size_t memTotalAllocated(void);
 extern void memDataInit(mem_type, const char *, size_t, int);
 extern void memCheckInit(void);
 
 /* MemPool */
 extern MemPool *memPoolCreate(const char *label, size_t obj_size);
 extern void memPoolDestroy(MemPool * pool);
+extern void memPoolNonZero(MemPool * p);
 extern void *memPoolAlloc(MemPool * pool);
 extern void memPoolFree(MemPool * pool, void *obj);
 extern int memPoolWasUsed(const MemPool * pool);
@@ -889,20 +923,24 @@ extern char *stmemNodeGet(mem_node *);
 /*
  * store.c
  */
-extern StoreEntry *new_StoreEntry(int, const char *, const char *);
+extern StoreEntry *new_StoreEntry(int, const char *);
+extern void storeEntrySetStoreUrl(StoreEntry * e, const char *store_url);
 extern StoreEntry *storeGet(const cache_key *);
 extern StoreEntry *storeGetPublic(const char *uri, const method_t method);
 extern StoreEntry *storeGetPublicByRequest(request_t * request);
 extern StoreEntry *storeGetPublicByRequestMethod(request_t * request, const method_t method);
-extern StoreEntry *storeCreateEntry(const char *, const char *, request_flags, method_t);
+extern StoreEntry *storeCreateEntry(const char *, request_flags, method_t);
 extern void storeSetPublicKey(StoreEntry *);
 extern void storeComplete(StoreEntry *);
 extern void storeInit(void);
 extern void storeAbort(StoreEntry *);
 extern void storeAppend(StoreEntry *, const char *, int);
-extern void storeLockObject(StoreEntry *);
+extern void storeLockObjectDebug(StoreEntry *, const char *file, const int line);
 extern void storeRelease(StoreEntry *);
-extern int storeUnlockObject(StoreEntry *);
+extern int storeUnlockObjectDebug(StoreEntry *, const char *file, const int line);
+extern const char *storeLookupUrl(const StoreEntry * e);
+#define	storeLockObject(a) storeLockObjectDebug(a, __FILE__, __LINE__);
+#define	storeUnlockObject(a) storeUnlockObjectDebug(a, __FILE__, __LINE__);
 extern EVH storeMaintainSwapSpace;
 extern void storeExpireNow(StoreEntry *);
 extern void storeReleaseRequest(StoreEntry *);
@@ -913,11 +951,11 @@ extern int expiresMoreThan(time_t, time_t);
 extern int storeEntryValidToSend(StoreEntry *);
 extern void storeTimestampsSet(StoreEntry *);
 extern void storeRegisterAbort(StoreEntry * e, STABH * cb, void *);
-extern void storeClientUnregisterAbort(StoreEntry * e);
+extern void storeUnregisterAbort(StoreEntry * e);
 extern void storeMemObjectDump(MemObject * mem);
 extern void storeEntryDump(const StoreEntry * e, int debug_lvl);
 extern const char *storeUrl(const StoreEntry *);
-extern void storeCreateMemObject(StoreEntry *, const char *, const char *);
+extern void storeCreateMemObject(StoreEntry *, const char *);
 extern void storeCopyNotModifiedReplyHeaders(MemObject * O, MemObject * N);
 extern void storeBuffer(StoreEntry *);
 extern void storeBufferFlush(StoreEntry *);
@@ -946,6 +984,9 @@ extern void storeReplAdd(const char *, REMOVALPOLICYCREATE *);
 void storeDeferRead(StoreEntry *, int fd);
 void storeResumeRead(StoreEntry *);
 void storeResetDefer(StoreEntry *);
+extern int memHaveHeaders(const MemObject * mem);
+extern void storeUpdate(StoreEntry * e, request_t *);
+
 
 /* store_modules.c */
 extern void storeFsSetup(void);
@@ -1065,6 +1106,7 @@ extern squid_off_t storeSwapOutObjectBytesOnDisk(const MemObject * mem);
  */
 extern store_client *storeClientRegister(StoreEntry * e, void *data);
 extern void storeClientCopy(store_client *, StoreEntry *, squid_off_t, squid_off_t, size_t, char *, STCB *, void *);
+extern void storeClientCopyHeaders(store_client *, StoreEntry *, STHCB *, void *);
 extern int storeClientCopyPending(store_client *, StoreEntry * e, void *data);
 extern int storeClientUnregister(store_client * sc, StoreEntry * e, void *data);
 extern squid_off_t storeLowestMemReaderOffset(const StoreEntry * entry);
@@ -1093,6 +1135,7 @@ extern void no_suid(void);
 extern void writePidFile(void);
 extern void setSocketShutdownLifetimes(int);
 extern void setMaxFD(void);
+extern void setSystemLimits(void);
 extern time_t getCurrentTime(void);
 extern int percent(int, int);
 extern double dpercent(double, double);
@@ -1121,7 +1164,7 @@ extern void unlinkdUnlink(const char *);
 extern char *url_convert_hex(char *org_url, int allocate);
 extern char *url_escape(const char *url);
 extern protocol_t urlParseProtocol(const char *);
-extern method_t urlParseMethod(const char *);
+extern method_t urlParseMethod(const char *, int len);
 extern void urlInitialize(void);
 extern request_t *urlParse(method_t, char *);
 extern const char *urlCanonical(request_t *);
@@ -1132,9 +1175,9 @@ extern int urlCheckRequest(const request_t *);
 extern int urlDefaultPort(protocol_t p);
 extern char *urlCanonicalClean(const request_t *);
 extern char *urlHostname(const char *url);
-extern void parse_extension_method(const char *(*methods)[]);
-extern void free_extension_method(const char *(*_methods)[]);
-extern void dump_extension_method(StoreEntry * entry, const char *name, const char **methods);
+extern void parse_extension_method(rms_t(*foo)[]);
+extern void free_extension_method(rms_t(*foo)[]);
+extern void dump_extension_method(StoreEntry * entry, const char *name, rms_t * methods);
 
 extern void useragentOpenLog(void);
 extern void useragentRotateLog(void);
@@ -1153,12 +1196,11 @@ extern void errorSend(int fd, ErrorState *);
 extern void errorAppendEntry(StoreEntry *, ErrorState *);
 extern void errorStateFree(ErrorState * err);
 extern int errorReservePageId(const char *page_name);
-extern ErrorState *errorCon(err_type type, http_status);
+extern ErrorState *errorCon(err_type type, http_status, request_t * request);
 extern int errorPageId(const char *page_name);
 
 extern void pconnPush(int, const char *host, u_short port, const char *domain, struct in_addr *client_address, u_short client_port);
-extern int pconnPop(const char *host, u_short port, const char *domain, struct in_addr *client_address, u_short client_port);
-extern hash_link *pconnLookup(const char *peer, u_short port, const char *domain, struct in_addr *client_address, u_short client_port);
+extern int pconnPop(const char *host, u_short port, const char *domain, struct in_addr *client_address, u_short client_port, int *idle);
 extern void pconnInit(void);
 
 extern int asnMatchIp(void *, struct in_addr);
@@ -1184,6 +1226,7 @@ extern int xrename(const char *from, const char *to);
 extern int isPowTen(int);
 extern void parseEtcHosts(void);
 extern int getMyPort(void);
+extern int parse_sockaddr(char *s, struct sockaddr_in *addr);
 
 char *strwordtok(char *buf, char **t);
 void strwordquote(MemBuf * mb, const char *str);
@@ -1224,12 +1267,13 @@ extern void stringAppend(String * s, const char *buf, int len);
 /*
  * ipc.c
  */
-extern int ipcCreate(int type,
+extern pid_t ipcCreate(int type,
     const char *prog,
     const char *const args[],
     const char *name,
     int *rfd,
-    int *wfd);
+    int *wfd,
+    void **hIpc);
 
 /* CacheDigest */
 extern CacheDigest *cacheDigestCreate(int capacity, int bpe);
@@ -1318,8 +1362,10 @@ extern void *leakFreeFL(void *, const char *, int);
 extern Logfile *logfileOpen(const char *path, size_t bufsz, int);
 extern void logfileClose(Logfile * lf);
 extern void logfileRotate(Logfile * lf);
-extern void logfileWrite(Logfile * lf, void *buf, size_t len);
+extern void logfileWrite(Logfile * lf, char *buf, size_t len);
 extern void logfileFlush(Logfile * lf);
+extern void logfileLineStart(Logfile * lf);
+extern void logfileLineEnd(Logfile * lf);
 #if STDC_HEADERS
 extern void
 logfilePrintf(Logfile * lf, const char *fmt,...) PRINTF_FORMAT_ARG2;
@@ -1351,7 +1397,7 @@ extern unsigned int url_checksum(const char *url);
  */
 extern StatCounters *snmpStatGet(int);
 
-/* CygWin & Windows NT Port */
+/* Cygwin & native Windows Port */
 /* win32.c */
 #ifdef _SQUID_WIN32_
 extern int WIN32_Subsystem_Init(int *, char ***);
@@ -1365,6 +1411,9 @@ extern int WIN32_pipe(int[2]);
 extern int WIN32_getrusage(int, struct rusage *);
 extern void WIN32_ExceptionHandlerInit(void);
 extern int SquidMain(int, char **);
+#ifdef _SQUID_MSWIN_
+extern DWORD WIN32_IpAddrChangeMonitorInit();
+#endif
 #endif
 
 /* external_acl.c */
@@ -1385,6 +1434,15 @@ extern char *strtokFile(void);
 const char *externalAclMessage(external_acl_entry * entry);
 
 
+/* refresh_check.c */
+extern void parse_refreshCheckHelper(refresh_check_helper **);
+extern void dump_refreshCheckHelper(StoreEntry * sentry, const char *name, const refresh_check_helper *);
+extern void free_refreshCheckHelper(refresh_check_helper **);
+extern void refreshCheckSubmit(StoreEntry * entry, REFRESHCHECK * callback, void *data);
+extern void refreshCheckInit(void);
+extern void refreshCheckConfigure(void);
+extern void refreshCheckShutdown(void);
+
 #if USE_WCCPv2
 extern void parse_wccp2_service(void *v);
 extern void free_wccp2_service(void *v);
@@ -1407,6 +1465,33 @@ extern int errorMapStart(const errormap * map, request_t * req, HttpReply * repl
 /* ETag support */
 void storeLocateVaryDone(VaryData * data);
 void storeLocateVary(StoreEntry * e, int offset, const char *vary_data, String accept_encoding, STLVCB * callback, void *cbdata);
-void storeAddVary(const char *url, const char *log_url, const method_t method, const cache_key * key, const char *etag, const char *vary, const char *vary_headers, const char *accept_encoding);
+void storeAddVary(const char *url, const method_t method, const cache_key * key, const char *etag, const char *vary, const char *vary_headers, const char *accept_encoding);
+
+/* New HTTP message parsing support */
+extern void HttpMsgBufInit(HttpMsgBuf * hmsg, const char *buf, size_t size);
+extern void httpMsgBufDone(HttpMsgBuf * hmsg);
+extern int httpMsgParseRequestLine(HttpMsgBuf * hmsg);
+extern int httpMsgParseRequestHeader(request_t * req, HttpMsgBuf * hmsg);
+extern int httpMsgFindHeadersEnd(HttpMsgBuf * hmsg);
+
+extern const char *xinet_ntoa(const struct in_addr addr);
+
+/* client_side.c */
+extern aclCheck_t *clientAclChecklistCreate(const acl_access * acl, const clientHttpRequest * http);
+extern void clientInterpretRequestHeaders(clientHttpRequest * http);
+extern void clientAccessCheck2(void *data);
+extern void clientFinishRewriteStuff(clientHttpRequest * http);
+
+
+/* client_side_redirect.c */
+extern void clientRedirectAccessCheckDone(int answer, void *data);
+extern void clientRedirectStart(clientHttpRequest * http);
+extern void clientRedirectDone(void *data, char *result);
+
+/* client_side_storeurl_rewrite.c */
+extern void clientStoreURLRewriteAccessCheckDone(int answer, void *data);
+extern void clientStoreURLRewriteStart(clientHttpRequest * http);
+extern void clientStoreURLRewriteDone(void *data, char *result);
+
 
 #endif /* SQUID_PROTOS_H */
