@@ -1,6 +1,6 @@
 
 /*
- * $Id: store.c,v 1.565 2006/07/17 02:32:00 hno Exp $
+ * $Id: store.c,v 1.584.2.3 2008/05/27 12:49:39 hno Exp $
  *
  * DEBUG: section 20    Storage Manager
  * AUTHOR: Harvest Derived
@@ -79,7 +79,7 @@ extern OBJH storeIOStats;
 static int storeEntryValidLength(const StoreEntry *);
 static void storeGetMemSpace(int);
 static void storeHashDelete(StoreEntry *);
-static MemObject *new_MemObject(const char *, const char *);
+static MemObject *new_MemObject(const char *);
 static void destroy_MemObject(StoreEntry *);
 static FREE destroy_StoreEntry;
 static void storePurgeMem(StoreEntry *);
@@ -100,18 +100,18 @@ unsigned int
 url_checksum(const char *url)
 {
     unsigned int ck;
-    MD5_CTX M;
+    SQUID_MD5_CTX M;
     static unsigned char digest[16];
-    MD5Init(&M);
-    MD5Update(&M, (unsigned char *) url, strlen(url));
-    MD5Final(digest, &M);
+    SQUID_MD5Init(&M);
+    SQUID_MD5Update(&M, (unsigned char *) url, strlen(url));
+    SQUID_MD5Final(digest, &M);
     xmemcpy(&ck, digest, sizeof(ck));
     return ck;
 }
 #endif
 
 static MemObject *
-new_MemObject(const char *url, const char *log_url)
+new_MemObject(const char *url)
 {
     MemObject *mem = memAllocate(MEM_MEMOBJECT);
     mem->reply = httpReplyCreate();
@@ -119,21 +119,32 @@ new_MemObject(const char *url, const char *log_url)
 #if URL_CHECKSUM_DEBUG
     mem->chksum = url_checksum(mem->url);
 #endif
-    mem->log_url = xstrdup(log_url);
     mem->object_sz = -1;
     mem->serverfd = -1;
-    /* XXX account log_url */
     debug(20, 3) ("new_MemObject: returning %p\n", mem);
     return mem;
 }
 
+int
+memHaveHeaders(const MemObject * mem)
+{
+    if (!mem)
+	return 0;
+    if (!mem->reply)
+	return 0;
+    if (mem->reply->pstate != psParsed)
+	return 0;
+    return 1;
+}
+
+
 StoreEntry *
-new_StoreEntry(int mem_obj_flag, const char *url, const char *log_url)
+new_StoreEntry(int mem_obj_flag, const char *url)
 {
     StoreEntry *e = NULL;
     e = memAllocate(MEM_STOREENTRY);
     if (mem_obj_flag)
-	e->mem_obj = new_MemObject(url, log_url);
+	e->mem_obj = new_MemObject(url);
     debug(20, 3) ("new_StoreEntry: returning %p\n", e);
     e->expires = e->lastmod = e->lastref = e->timestamp = -1;
     e->swap_filen = -1;
@@ -141,6 +152,16 @@ new_StoreEntry(int mem_obj_flag, const char *url, const char *log_url)
     return e;
 }
 
+void
+storeEntrySetStoreUrl(StoreEntry * e, const char *store_url)
+{
+    /* XXX eww, another strdup! */
+    if (!e->mem_obj)
+	return;
+    safe_free(e->mem_obj->store_url);
+    e->mem_obj->store_url = xstrdup(store_url);
+}
+
 static void
 destroy_MemObject(StoreEntry * e)
 {
@@ -166,12 +187,15 @@ destroy_MemObject(StoreEntry * e)
 	storeUnlockObject(mem->ims_entry);
 	mem->ims_entry = NULL;
     }
+    if (mem->old_entry) {
+	storeUnlockObject(mem->old_entry);
+	mem->old_entry = NULL;
+    }
     httpReplyDestroy(mem->reply);
     requestUnlink(mem->request);
     mem->request = NULL;
     ctx_exit(ctx);		/* must exit before we free mem->url */
     safe_free(mem->url);
-    safe_free(mem->log_url);	/* XXX account log_url */
     safe_free(mem->vary_headers);
     safe_free(mem->vary_encoding);
     memFree(mem, MEM_MEMOBJECT);
@@ -264,10 +288,10 @@ storeEntryDereferenced(StoreEntry * e)
 }
 
 void
-storeLockObject(StoreEntry * e)
+storeLockObjectDebug(StoreEntry * e, const char *file, const int line)
 {
     e->lock_count++;
-    debug(20, 3) ("storeLockObject: key '%s' count=%d\n",
+    debug(20, 3) ("storeLockObject: (%s:%d): key '%s' count=%d\n", file, line,
 	storeKeyText(e->hash.key), (int) e->lock_count);
     e->lastref = squid_curtime;
     storeEntryReferenced(e);
@@ -292,10 +316,10 @@ storeReleaseRequest(StoreEntry * e)
 /* unlock object, return -1 if object get released after unlock
  * otherwise lock_count */
 int
-storeUnlockObject(StoreEntry * e)
+storeUnlockObjectDebug(StoreEntry * e, const char *file, const int line)
 {
     e->lock_count--;
-    debug(20, 3) ("storeUnlockObject: key '%s' count=%d\n",
+    debug(20, 3) ("storeUnlockObject: (%s:%d): key '%s' count=%d\n", file, line,
 	storeKeyText(e->hash.key), e->lock_count);
     if (e->lock_count)
 	return (int) e->lock_count;
@@ -418,18 +442,21 @@ free_AddVaryState(void *data)
 {
     AddVaryState *state = data;
     debug(11, 2) ("free_AddVaryState: %p\n", data);
-    if (!state->done && state->key) {
-	storeAppendPrintf(state->e, "Key: %s\n", state->key);
-	if (state->accept_encoding)
-	    storeAppendPrintf(state->e, "Accept-Encoding: %s\n", state->accept_encoding);
-	if (state->etag)
-	    storeAppendPrintf(state->e, "ETag: %s\n", state->etag);
-	storeAppendPrintf(state->e, "VaryData: %s\n", state->vary_headers);
+    if (!EBIT_TEST(state->e->flags, ENTRY_ABORTED)) {
+	storeBuffer(state->e);
+	if (!state->done && state->key) {
+	    storeAppendPrintf(state->e, "Key: %s\n", state->key);
+	    if (state->accept_encoding)
+		storeAppendPrintf(state->e, "Accept-Encoding: %s\n", state->accept_encoding);
+	    if (state->etag)
+		storeAppendPrintf(state->e, "ETag: %s\n", state->etag);
+	    storeAppendPrintf(state->e, "VaryData: %s\n", state->vary_headers);
+	}
+	storeTimestampsSet(state->e);
+	storeComplete(state->e);
+	storeTimestampsSet(state->e);
+	storeBufferFlush(state->e);
     }
-    storeBufferFlush(state->e);
-    storeTimestampsSet(state->e);
-    storeComplete(state->e);
-    storeTimestampsSet(state->e);
     storeUnlockObject(state->e);
     state->e = NULL;
     if (state->sc) {
@@ -539,6 +566,12 @@ storeAddVaryReadOld(void *data, char *buf, ssize_t size)
 	cbdataFree(state);
 	return;
     }
+    if (EBIT_TEST(state->e->flags, ENTRY_ABORTED)) {
+	debug(11, 1) ("storeAddVaryReadOld: New index aborted at %d (%d)\n", (int) state->seen_offset, (int) size);
+	cbdataFree(state);
+	return;
+    }
+    storeBuffer(state->e);
     if (state->seen_offset != 0) {
 	state->seen_offset = state->seen_offset + size;
     } else {
@@ -645,7 +678,10 @@ storeAddVaryReadOld(void *data, char *buf, ssize_t size)
 	    break;
 	assert(p <= (buf + size));
     }
-    if (p == state->buf && size == state->buf_size) {
+    state->buf_offset = l;
+    if (l && p != state->buf)
+	memmove(state->buf, p, l);
+    if (state->buf_offset == state->buf_size) {
 	/* Oops.. the buffer size is not sufficient. Grow */
 	if (state->buf_size < 65536) {
 	    debug(11, 2) ("storeAddVaryReadOld: Increasing entry buffer size to %d\n", (int) state->buf_size * 2);
@@ -657,10 +693,8 @@ storeAddVaryReadOld(void *data, char *buf, ssize_t size)
 	    return;
 	}
     }
-    state->buf_offset = l;
-    if (l)
-	memmove(state->buf, p, l);
     debug(11, 3) ("storeAddVaryReadOld: %p seen_offset=%" PRINTF_OFF_T " buf_offset=%d\n", data, state->seen_offset, (int) state->buf_offset);
+    storeBufferFlush(state->e);
     storeClientCopy(state->sc, state->oe,
 	state->seen_offset,
 	state->seen_offset,
@@ -676,10 +710,9 @@ storeAddVaryReadOld(void *data, char *buf, ssize_t size)
  * At leas one of key or etag must be specified, preferably both.
  */
 void
-storeAddVary(const char *url, const char *log_url, const method_t method, const cache_key * key, const char *etag, const char *vary, const char *vary_headers, const char *accept_encoding)
+storeAddVary(const char *url, const method_t method, const cache_key * key, const char *etag, const char *vary, const char *vary_headers, const char *accept_encoding)
 {
     AddVaryState *state;
-    http_version_t version;
     request_flags flags = null_request_flags;
     CBDATA_INIT_TYPE_FREECB(AddVaryState, free_AddVaryState);
     state = cbdataAlloc(AddVaryState);
@@ -697,14 +730,12 @@ storeAddVary(const char *url, const char *log_url, const method_t method, const
     if (state->oe)
 	storeLockObject(state->oe);
     flags.cachable = 1;
-    state->e = storeCreateEntry(url, log_url, flags, method);
-    httpBuildVersion(&version, 1, 0);
-    httpReplySetHeaders(state->e->mem_obj->reply, version, HTTP_OK, "Internal marker object", "x-squid-internal/vary", -1, -1, squid_curtime + 100000);
+    state->e = storeCreateEntry(url, flags, method);
+    httpReplySetHeaders(state->e->mem_obj->reply, HTTP_OK, "Internal marker object", "x-squid-internal/vary", -1, -1, squid_curtime + 100000);
     httpHeaderPutStr(&state->e->mem_obj->reply->header, HDR_VARY, vary);
     storeSetPublicKey(state->e);
     storeBuffer(state->e);
     httpReplySwapOut(state->e->mem_obj->reply, state->e);
-    storeBufferFlush(state->e);
     if (state->oe) {
 	/* Here we need to tack on the old etag/vary information, and we should
 	 * merge, clean up etc
@@ -722,7 +753,7 @@ storeAddVary(const char *url, const char *log_url, const method_t method, const
 	 */
 	/* Swap in the dummy Vary object */
 	if (!state->oe->mem_obj) {
-	    storeCreateMemObject(state->oe, state->url, log_url);
+	    storeCreateMemObject(state->oe, state->url);
 	    state->oe->mem_obj->method = method;
 	}
 	state->sc = storeClientRegister(state->oe, state);
@@ -810,7 +841,7 @@ storeLocateVaryCallback(LocateVaryState * state)
 	state->buf = NULL;
     }
     cbdataFree(state);
-    debug(11, 2) ("storeLocateVary: DONE\n");
+    debug(11, 2) ("storeLocateVaryCallback: DONE\n");
 }
 
 static void
@@ -936,7 +967,7 @@ storeLocateVary(StoreEntry * e, int offset, const char *vary_data, String accept
     state->buf = memAllocBuf(4096, &state->buf_size);
     state->sc = storeClientRegister(state->e, state);
     state->seen_offset = offset;
-    if (strCmp(e->mem_obj->reply->content_type, "x-squid-internal/vary") != 0) {
+    if (!strLen(e->mem_obj->reply->content_type) || strCmp(e->mem_obj->reply->content_type, "x-squid-internal/vary") != 0) {
 	/* This is not our Vary marker object. Bail out. */
 	debug(33, 1) ("storeLocateVary: Not our vary marker object, %s = '%s', '%s'/'%s'\n",
 	    storeKeyText(e->hash.key), e->mem_obj->url, vary_data, strBuf(accept_encoding) ? strBuf(accept_encoding) : "-");
@@ -958,8 +989,14 @@ storeSetPublicKey(StoreEntry * e)
     StoreEntry *e2 = NULL;
     const cache_key *newkey;
     MemObject *mem = e->mem_obj;
-    if (e->hash.key && !EBIT_TEST(e->flags, KEY_PRIVATE))
-	return;			/* is already public */
+    if (e->hash.key && !EBIT_TEST(e->flags, KEY_PRIVATE)) {
+	if (EBIT_TEST(e->flags, KEY_EARLY_PUBLIC)) {
+	    EBIT_CLR(e->flags, KEY_EARLY_PUBLIC);
+	    storeSetPrivateKey(e);	/* wasn't really public yet, reset the key */
+	} else {
+	    return;		/* is already public */
+	}
+    }
     assert(mem);
     /*
      * We can't make RELEASE_REQUEST objects public.  Depending on
@@ -989,19 +1026,21 @@ storeSetPublicKey(StoreEntry * e)
 		 * to record the new variance key
 		 */
 		safe_free(request->vary_headers);	/* free old "bad" variance key */
-		pe = storeGetPublic(mem->url, mem->method);
+		pe = storeGetPublic(storeLookupUrl(e), mem->method);
 		if (pe)
 		    storeRelease(pe);
 	    }
 	    /* Make sure the request knows the variance status */
-	    if (!request->vary_headers) {
-		const char *vary = httpMakeVaryMark(request, mem->reply);
-		if (vary)
-		    request->vary_headers = xstrdup(vary);
+	    else if (!request->vary_headers) {
+		if (!httpMakeVaryMark(request, mem->reply)) {
+		    /* Release the object if we could not index the variance */
+		    storeReleaseRequest(e);
+		    return;
+		}
 	    }
 	}
 	newkey = storeKeyPublicByRequest(mem->request);
-	if (mem->vary_headers) {
+	if (mem->vary_headers && !EBIT_TEST(e->flags, KEY_EARLY_PUBLIC)) {
 	    String vary = StringNull;
 	    String varyhdr;
 	    varyhdr = httpHeaderGetList(&mem->reply->header, HDR_VARY);
@@ -1015,11 +1054,11 @@ storeSetPublicKey(StoreEntry * e)
 		strListAdd(&vary, strBuf(varyhdr), ',');
 	    stringClean(&varyhdr);
 #endif
-	    storeAddVary(mem->url, mem->log_url, mem->method, newkey, httpHeaderGetStr(&mem->reply->header, HDR_ETAG), strBuf(vary), mem->vary_headers, mem->vary_encoding);
+	    storeAddVary(mem->url, mem->method, newkey, httpHeaderGetStr(&mem->reply->header, HDR_ETAG), strBuf(vary), mem->vary_headers, mem->vary_encoding);
 	    stringClean(&vary);
 	}
     } else {
-	newkey = storeKeyPublic(mem->url, mem->method);
+	newkey = storeKeyPublic(storeLookupUrl(e), mem->method);
     }
     if ((e2 = (StoreEntry *) hash_lookup(store_table, newkey))) {
 	debug(20, 3) ("storeSetPublicKey: Making old '%s' private.\n", mem->url);
@@ -1028,7 +1067,7 @@ storeSetPublicKey(StoreEntry * e)
 	if (mem->request)
 	    newkey = storeKeyPublicByRequest(mem->request);
 	else
-	    newkey = storeKeyPublic(mem->url, mem->method);
+	    newkey = storeKeyPublic(storeLookupUrl(e), mem->method);
     }
     if (e->hash.key)
 	storeHashDelete(e);
@@ -1039,13 +1078,13 @@ storeSetPublicKey(StoreEntry * e)
 }
 
 StoreEntry *
-storeCreateEntry(const char *url, const char *log_url, request_flags flags, method_t method)
+storeCreateEntry(const char *url, request_flags flags, method_t method)
 {
     StoreEntry *e = NULL;
     MemObject *mem = NULL;
     debug(20, 3) ("storeCreateEntry: '%s'\n", url);
 
-    e = new_StoreEntry(STORE_ENTRY_WITH_MEMOBJ, url, log_url);
+    e = new_StoreEntry(STORE_ENTRY_WITH_MEMOBJ, url);
     e->lock_count = 1;		/* Note lock here w/o calling storeLock() */
     mem = e->mem_obj;
     mem->method = method;
@@ -1291,6 +1330,11 @@ storeComplete(StoreEntry * e)
     if (e->mem_obj->request)
 	e->mem_obj->request->hier.store_complete_stop = current_time;
 #endif
+    e->mem_obj->refresh_timestamp = 0;
+    if (e->mem_obj->old_entry) {
+	if (e->mem_obj->old_entry->mem_obj)
+	    e->mem_obj->old_entry->mem_obj->refresh_timestamp = 0;
+    }
     /*
      * We used to call InvokeHandlers, then storeSwapOut.  However,
      * Madhukar Reddy <myreddy@persistence.com> reported that
@@ -1314,7 +1358,7 @@ storeAbort(StoreEntry * e)
     assert(mem != NULL);
     debug(20, 6) ("storeAbort: %s\n", storeKeyText(e->hash.key));
     storeLockObject(e);		/* lock while aborting */
-    storeNegativeCache(e);
+    storeExpireNow(e);
     storeReleaseRequest(e);
     EBIT_SET(e->flags, ENTRY_ABORTED);
     storeSetMemStatus(e, NOT_IN_MEMORY);
@@ -1356,14 +1400,17 @@ storeGetMemSpace(int size)
     pages_needed = (size / SM_PAGE_SIZE) + 1;
     if (memInUse(MEM_MEM_NODE) + pages_needed < store_pages_max)
 	return;
-    debug(20, 2) ("storeGetMemSpace: Starting, need %d pages\n", pages_needed);
+    debug(20, 3) ("storeGetMemSpace: Starting, need %d pages\n", pages_needed);
     /* XXX what to set as max_scan here? */
     walker = mem_policy->PurgeInit(mem_policy, 100000);
     while ((e = walker->Next(walker))) {
+	debug(20, 3) ("storeGetMemSpace: purging %p\n", e);
 	storePurgeMem(e);
 	released++;
-	if (memInUse(MEM_MEM_NODE) + pages_needed < store_pages_max)
+	if (memInUse(MEM_MEM_NODE) + pages_needed < store_pages_max) {
+	    debug(20, 3) ("storeGetMemSpace: we finally have enough free memory!\n");
 	    break;
+	}
     }
     walker->Done(walker);
     debug(20, 3) ("storeGetMemSpace stats:\n");
@@ -1600,7 +1647,37 @@ storeKeepInMemory(const StoreEntry * e)
 void
 storeNegativeCache(StoreEntry * e)
 {
-    e->expires = squid_curtime + Config.negativeTtl;
+    StoreEntry *oe = e->mem_obj->old_entry;
+    time_t expires = e->expires;
+    refresh_cc cc = refreshCC(e, e->mem_obj->request);
+    if (expires == -1)
+	expires = squid_curtime + cc.negative_ttl;
+    if (oe && !EBIT_TEST(oe->flags, KEY_PRIVATE) && !EBIT_TEST(oe->flags, ENTRY_REVALIDATE)) {
+	HttpHdrCc *oldcc = oe->mem_obj->reply->cache_control;
+	if (oldcc && EBIT_TEST(oldcc->mask, CC_STALE_IF_ERROR) && oldcc->stale_if_error >= 0)
+	    cc.max_stale = oldcc->stale_if_error;
+	if (cc.max_stale >= 0) {
+	    time_t max_expires;
+	    storeTimestampsSet(oe);
+	    max_expires = oe->expires + cc.max_stale;
+	    /* Bail out if beyond the stale-if-error staleness limit */
+	    if (max_expires <= squid_curtime)
+		goto cache_error_response;
+	    /* Limit expiry time to stale-if-error/max_stale */
+	    if (expires > max_expires)
+		expires = max_expires;
+	}
+	/* Block the new error from getting cached */
+	EBIT_CLR(e->flags, ENTRY_CACHABLE);
+	/* And negatively cache the old one */
+	if (oe->expires < expires)
+	    oe->expires = expires;
+	EBIT_SET(oe->flags, REFRESH_FAILURE);
+	return;
+    }
+  cache_error_response:
+    if (e->expires < expires)
+	e->expires = expires;
     EBIT_SET(e->flags, ENTRY_NEGCACHED);
 }
 
@@ -1636,9 +1713,12 @@ storeEntryValidToSend(StoreEntry * e)
     if (EBIT_TEST(e->flags, ENTRY_ABORTED))
 	return 0;
     /* Entries which seem to have got stuck is not valid to send to new clients */
-    if (e->store_status == STORE_PENDING)
+    if (e->store_status == STORE_PENDING) {
 	if (!e->mem_obj || e->mem_obj->refresh_timestamp + 30 < squid_curtime)
 	    return 0;
+	else
+	    return -1;
+    }
     return 1;
 }
 
@@ -1668,7 +1748,10 @@ storeTimestampsSet(StoreEntry * entry)
     if (age > squid_curtime - served_date)
 	if (squid_curtime > age)
 	    served_date = squid_curtime - age;
-    entry->expires = reply->expires;
+    if (reply->expires > 0 && reply->date > -1)
+	entry->expires = served_date + (reply->expires - reply->date);
+    else
+	entry->expires = reply->expires;
     entry->lastmod = reply->last_modified;
     entry->timestamp = served_date;
 }
@@ -1684,7 +1767,7 @@ storeRegisterAbort(StoreEntry * e, STABH * cb, void *data)
 }
 
 void
-storeClientUnregisterAbort(StoreEntry * e)
+storeUnregisterAbort(StoreEntry * e)
 {
     MemObject *mem = e->mem_obj;
     assert(mem);
@@ -1713,9 +1796,9 @@ storeMemObjectDump(MemObject * mem)
 	mem->reply);
     debug(20, 1) ("MemObject->request: %p\n",
 	mem->request);
-    debug(20, 1) ("MemObject->log_url: %p %s\n",
-	mem->log_url,
-	checkNullString(mem->log_url));
+    debug(20, 1) ("MemObject->url: %p %s\n",
+	mem->url,
+	checkNullString(mem->url));
 }
 
 void
@@ -1786,12 +1869,25 @@ storeUrl(const StoreEntry * e)
 	return e->mem_obj->url;
 }
 
+const char *
+storeLookupUrl(const StoreEntry * e)
+{
+    if (e == NULL)
+	return "[null_entry]";
+    else if (e->mem_obj == NULL)
+	return "[null_mem_obj]";
+    else if (e->mem_obj->store_url)
+	return e->mem_obj->store_url;
+    else
+	return e->mem_obj->url;
+}
+
 void
-storeCreateMemObject(StoreEntry * e, const char *url, const char *log_url)
+storeCreateMemObject(StoreEntry * e, const char *url)
 {
     if (e->mem_obj)
 	return;
-    e->mem_obj = new_MemObject(url, log_url);
+    e->mem_obj = new_MemObject(url);
 }
 
 /* this just sets DELAY_SENDING */
